import os
import ssl
import anyio
import logging

from anyio import run as _anyio_run

log = logging.getLogger(__name__)

WantRead = (BlockingIOError, InterruptedError, ssl.SSLWantReadError)
WantWrite = (BlockingIOError, InterruptedError, ssl.SSLWantWriteError)


def _new_run(func, *args, backend=None, backend_options=None):
    if backend is None:
        backend = os.getenv("PURERPC_BACKEND", "asyncio")
    log.info("Selected {} backend".format(backend))
    _anyio_run(func, *args, backend=backend, backend_options=backend_options)


async def _new_sendall(self, data, *, flags=0):
    with memoryview(data).cast('B') as buffer:
        total_sent = 0
        while buffer:
            await self._check_cancelled()
            try:
                num_sent = self._raw_socket.send(buffer, flags)
                total_sent += num_sent
                buffer = buffer[num_sent:]
            except WantWrite:
                await self._wait_writable()
            except WantRead:  # pragma: no cover
                await self._wait_readable()
            except ssl.SSLEOFError:
                self._raw_socket.close()
                raise


def apply_monkeypatch():
    """Apply AnyIO monkeypatches (should merge upstream)"""
    anyio.run = _new_run
    anyio._networking.BaseSocket.sendall = _new_sendall


import base64
import collections


class HeaderDict(collections.OrderedDict):
    def __init__(self, values):
        super().__init__()
        for key, value in values:
            if key not in self:
                self[key] = [value]
            else:
                self[key].append(value)
        for key in self:
            if len(self[key]) == 1:
                self[key] = self[key][0]

    def extract_headers(self, header_name: str):
        """Returns all headers with name == header_name as list of tuples (name, value)"""
        if header_name.startswith("grpc-"):
            return ()
        else:
            value = self.pop(header_name)
            is_binary = header_name.endswith("-bin")

            if not isinstance(value, list):
                value_list = [value]
            else:
                value_list = value

            if is_binary:
                return ((header_name, b64decode(value)) for value_sublist in value_list for value
                        in value_sublist.split(","))
            else:
                return ((header_name, value) for value in value_list)


def sanitize_headers(headers):
    for name, value in headers:
        if isinstance(value, bytes) and not name.endswith("-bin"):
            raise ValueError("Got binary value for header name '{}', but name does not end "
                             "with '-bin' suffix".format(name))
        if name.startswith("grpc-"):
            raise ValueError("Got header with name '{}', but custom metadata headers should "
                             "not start with 'grpc-' prefix".format(name))
        if name.endswith("-bin"):
            yield name, b64encode(value)
        else:
            yield name, value


def b64decode(data: str) -> bytes:
    # Apply missing padding
    missing_padding = len(data) % 4
    if missing_padding:
        data += "=" * (4 - missing_padding)
    return base64.b64decode(data)


def b64encode(data: bytes) -> str:
    return base64.b64encode(data).rstrip(b"=").decode("utf-8")



import platform
import anyio
import pdb
import math


def is_linux():
    return platform.system() == "Linux"


def is_darwin():
    return platform.system() == "Darwin"


def get_linux_kernel_version():
    if not is_linux():
        return None
    release = platform.release()
    if not release:
        return None
    return tuple(map(int, release.split("-")[0].split(".")))


async def print_memory_growth_statistics(interval_sec=10.0, set_pdb_trace_every=math.inf):
    num_iters = 0
    import objgraph
    while True:
        num_iters += 1
        await anyio.sleep(interval_sec)
        objgraph.show_growth()
        if num_iters == set_pdb_trace_every:
            pdb.set_trace()
            num_iters = 0



import datetime

from purerpc.grpclib.config import GRPCConfiguration
from purerpc.grpclib.events import MessageReceived, RequestEnded, ResponseEnded

from .grpc_socket import GRPCStream, GRPCSocket


class GRPCProtoStream(GRPCStream):
    def expect_message_type(self, message_type):
        self._incoming_message_type = message_type

    async def send_message(self, message):
        return await super()._send(message.SerializeToString())

    async def receive_event(self):
        event = await super()._receive()
        if isinstance(event, MessageReceived) and hasattr(self, "_incoming_message_type"):
            binary_data = event.data
            event.data = self._incoming_message_type()
            event.data.ParseFromString(binary_data)
        return event

    async def receive_message(self):
        event = await self.receive_event()
        if isinstance(event, RequestEnded) or isinstance(event, ResponseEnded):
            return None
        elif isinstance(event, MessageReceived):
            return event.data
        else:
            return await self.receive_message()

    async def start_response(self, content_type_suffix="", custom_metadata=()):
        return await super().start_response(
            content_type_suffix if content_type_suffix else "+proto",
            custom_metadata)


class GRPCProtoSocket(GRPCSocket):
    StreamClass = GRPCProtoStream

    async def start_request(self, scheme: str, service_name: str, method_name: str,
                            message_type=None, authority=None, timeout: datetime.timedelta = None,
                            content_type_suffix="", custom_metadata=()):
        return await super().start_request(
            scheme, service_name, method_name, message_type, authority, timeout,
            content_type_suffix if content_type_suffix else "+proto", custom_metadata
        )



import enum
import typing
import collections


Stream = typing.AsyncIterator


class Cardinality(enum.Enum):
    UNARY_UNARY = 0
    UNARY_STREAM = 1
    STREAM_UNARY = 2
    STREAM_STREAM = 3

    @staticmethod
    def get_cardinality_for(*, request_stream, response_stream):
        if request_stream and response_stream:
            return Cardinality.STREAM_STREAM
        elif request_stream and not response_stream:
            return Cardinality.STREAM_UNARY
        elif not request_stream and response_stream:
            return Cardinality.UNARY_STREAM
        else:
            return Cardinality.UNARY_UNARY


class RPCSignature:
    def __init__(self, cardinality: Cardinality, request_type, response_type):
        self._cardinality = cardinality
        self._request_type = request_type
        self._response_type = response_type

    @property
    def cardinality(self):
        return self._cardinality

    @property
    def request_type(self):
        return self._request_type

    @property
    def response_type(self):
        return self._response_type

    @staticmethod
    def from_annotations(request_annotation, response_annotation):
        if (hasattr(request_annotation, "__origin__") and
                issubclass(request_annotation.__origin__, collections.abc.AsyncIterator)):
            request_type = request_annotation.__args__[0]
            request_stream = True
        else:
            request_type = request_annotation
            request_stream = False

        if (hasattr(response_annotation, "__origin__") and
                issubclass(response_annotation.__origin__, collections.abc.AsyncIterator)):
            response_type = response_annotation.__args__[0]
            response_stream = True
        else:
            response_type = response_annotation
            response_stream = False
        cardinality = Cardinality.get_cardinality_for(request_stream=request_stream,
                                                      response_stream=response_stream)
        return RPCSignature(cardinality, request_type, response_type)



import logging
import datetime

import h2.stream
import h2.errors
import h2.events
import h2.connection
import h2.exceptions
from h2.settings import SettingCodes

from .headers import HeaderDict, sanitize_headers
from .status import Status
from .config import GRPCConfiguration
from .events import MessageReceived, RequestReceived, RequestEnded, ResponseReceived, ResponseEnded
from .exceptions import ProtocolError
from .buffers import MessageReadBuffer, MessageWriteBuffer

logger = logging.getLogger(__name__)


def monkey_patch_h2_reset_on_closed_stream_bug():
    """
    H2 raises h2.exceptions.StreamClosedError when receiving RST_STREAM on a closed stream that was
    not closed by endpoint, for example:
    
    -> HEADERS frame
    -> DATA frame with END_STREAM flag set  # stream is now half-closed (local)
    <- HEADERS frame
    <- DATA frame
    <- HEADERS frame with END_STREAM flag set  # stream is now closed
    <- RST_STREAM

    This is the case when calling unary-unary requests with purerpc client and grpcio server
    (it sends END_STREAM + RST_STREAM because the RPC call is unary-unary and it cannot receive
    any more data).

    To fix this, we use a workaround.
    """
    h2.stream._transitions[h2.stream.StreamState.CLOSED,
                           h2.stream.StreamInputs.RECV_RST_STREAM] = \
        (None, h2.stream.StreamState.CLOSED)

    h2.stream._transitions[h2.stream.StreamState.CLOSED,
                           h2.stream.StreamInputs.RECV_WINDOW_UPDATE] = \
        (None, h2.stream.StreamState.CLOSED)


monkey_patch_h2_reset_on_closed_stream_bug()


class GRPCConnection:
    MAX_HEADER_LIST_SIZE = 8192
    MAX_CONCURRENT_STREAMS = 2 ** 16
    MAX_INBOUND_FRAME_SIZE = 2 ** 24 - 1
    MAX_CONCURRENT_STREAMS = 1000

    def __init__(self, config: GRPCConfiguration):
        self.config = config
        self.h2_connection = h2.connection.H2Connection(config._h2_config)
        self._monkey_patch_h2_connection()
        self.message_read_buffers = {}

    def _monkey_patch_h2_connection(self):
        self.h2_connection.local_settings = h2.settings.Settings(
            client=self.config.client_side,
            initial_values={
                SettingCodes.MAX_CONCURRENT_STREAMS: self.MAX_CONCURRENT_STREAMS,
                SettingCodes.INITIAL_WINDOW_SIZE: 2 * self.config.max_message_length,
                SettingCodes.MAX_FRAME_SIZE: self.MAX_INBOUND_FRAME_SIZE,
                SettingCodes.MAX_HEADER_LIST_SIZE: self.MAX_HEADER_LIST_SIZE,
            }
        )
        self.h2_connection.max_inbound_frame_size = self.MAX_INBOUND_FRAME_SIZE

    def _request_received(self, event: h2.events.RequestReceived):
        if event.stream_ended:
            raise ProtocolError("Stream ended before data was sent")
        request = RequestReceived.parse_from_stream_id_and_headers_destructive(
            event.stream_id, HeaderDict(event.headers))
        self.message_read_buffers[event.stream_id] = MessageReadBuffer(request.message_encoding,
            self.config.max_message_length)
        return [request]

    def _response_received(self, event: h2.events.ResponseReceived):
        headers = HeaderDict(event.headers)
        response_received = ResponseReceived.parse_from_stream_id_and_headers_destructive(
            event.stream_id, headers)
        if event.stream_ended:
            response_ended = ResponseEnded.parse_from_stream_id_and_headers_destructive(
                event.stream_id, headers)
            return [response_received, response_ended]
        else:
            if len(headers) > 0:
                raise ProtocolError("Unparsed headers: {}".format(headers))
            self.message_read_buffers[event.stream_id] = MessageReadBuffer(
                response_received.message_encoding, self.config.max_message_length)
            return [response_received]

    def _trailers_received(self, event: h2.events.TrailersReceived):
        response_ended = ResponseEnded.parse_from_stream_id_and_headers_destructive(
            event.stream_id, HeaderDict(event.headers))
        return [response_ended]

    def _informational_response_received(self, event: h2.events.InformationalResponseReceived):
        return []

    def _data_received(self, event: h2.events.DataReceived):
        try:
            self.message_read_buffers[event.stream_id].data_received(event.data,
                                                                     event.flow_controlled_length)
        except KeyError:
            self.h2_connection.reset_stream(event.stream_id, h2.errors.ErrorCodes.PROTOCOL_ERROR)

        iterator = (self.message_read_buffers[event.stream_id]
                    .read_all_complete_messages_flowcontrol())
        events = []
        for message, flow_controlled_length in iterator:
            events.append(MessageReceived(event.stream_id, message, flow_controlled_length))
        return events

    def _window_updated(self, event: h2.events.WindowUpdated):
        return [event]

    def _remote_settings_changed(self, event: h2.events.RemoteSettingsChanged):
        fake_event = h2.events.WindowUpdated()
        fake_event.stream_id = 0
        fake_event.delta = 1
        return [fake_event]

    def _ping_acknowledged(self, event: h2.events.PingAcknowledged):
        return []

    def _stream_ended(self, event: h2.events.StreamEnded):
        if event.stream_id in self.message_read_buffers:
            del self.message_read_buffers[event.stream_id]
            return [RequestEnded(event.stream_id)] if not self.config.client_side else []
        return []

    def _stream_reset(self, event: h2.events.StreamReset):
        return []

    def _push_stream_received(self, event: h2.events.PushedStreamReceived):
        return []

    def _settings_acknowledged(self, event: h2.events.SettingsAcknowledged):
        # TODO: implement this
        return []

    def _priority_updated(self, event: h2.events.PriorityUpdated):
        return []

    def _connection_terminated(self, event: h2.events.ConnectionTerminated):
        return []

    def _alternative_service_available(self, event: h2.events.AlternativeServiceAvailable):
        return []

    def _unknown_frame_received(self, event: h2.events.UnknownFrameReceived):
        return []

    def initiate_connection(self):
        self.h2_connection.initiate_connection()
        self.h2_connection.increment_flow_control_window(2 ** 30)

    def data_to_send(self, amount: int = None):
        return self.h2_connection.data_to_send(amount)

    def receive_data(self, data: bytes):
        events = self.h2_connection.receive_data(data)
        grpc_events = []
        for event in events:
            if isinstance(event, h2.events.RequestReceived):
                grpc_events.extend(self._request_received(event))
            elif isinstance(event, h2.events.ResponseReceived):
                grpc_events.extend(self._response_received(event))
            elif isinstance(event, h2.events.TrailersReceived):
                grpc_events.extend(self._trailers_received(event))
            elif isinstance(event, h2.events.InformationalResponseReceived):
                grpc_events.extend(self._informational_response_received(event))
            elif isinstance(event, h2.events.DataReceived):
                grpc_events.extend(self._data_received(event))
            elif isinstance(event, h2.events.WindowUpdated):
                grpc_events.extend(self._window_updated(event))
            elif isinstance(event, h2.events.RemoteSettingsChanged):
                grpc_events.extend(self._remote_settings_changed(event))
            elif isinstance(event, h2.events.PingAcknowledged):
                grpc_events.extend(self._ping_acknowledged(event))
            elif isinstance(event, h2.events.StreamEnded):
                grpc_events.extend(self._stream_ended(event))
            elif isinstance(event, h2.events.StreamReset):
                grpc_events.extend(self._stream_reset(event))
            elif isinstance(event, h2.events.PushedStreamReceived):
                grpc_events.extend(self._push_stream_received(event))
            elif isinstance(event, h2.events.SettingsAcknowledged):
                grpc_events.extend(self._settings_acknowledged(event))
            elif isinstance(event, h2.events.PriorityUpdated):
                grpc_events.extend(self._priority_updated(event))
            elif isinstance(event, h2.events.ConnectionTerminated):
                grpc_events.extend(self._connection_terminated(event))
            elif isinstance(event, h2.events.AlternativeServiceAvailable):
                grpc_events.extend(self._alternative_service_available(event))
            elif isinstance(event, h2.events.UnknownFrameReceived):
                grpc_events.extend(self._unknown_frame_received(event))

        return grpc_events

    def flow_control_window(self, stream_id: int):
        return min(self.h2_connection.max_outbound_frame_size,
                   self.h2_connection.local_flow_control_window(stream_id))

    def reset_stream(self, stream_id: int, error_code: h2.errors.ErrorCodes):
        self.h2_connection.reset_stream(stream_id, error_code)

    def acknowledge_received_data(self, stream_id: int, flow_controlled_length: int):
        self.h2_connection.acknowledge_received_data(flow_controlled_length, stream_id)

    def send_data(self, stream_id: int, data: bytes, end_stream: bool = False):
        self.h2_connection.send_data(stream_id, data, end_stream=end_stream)

    def get_next_available_stream_id(self):
        return self.h2_connection.get_next_available_stream_id()

    def start_request(self, stream_id: int, scheme: str, service_name: str, method_name: str,
                      message_type=None, authority=None, timeout: datetime.timedelta=None,
                      content_type_suffix="", custom_metadata=()):
        headers = [
            (":method", "POST"),
            (":scheme", scheme),
            (":path", "/{}/{}".format(service_name, method_name)),
            ("te", "trailers"),
            ("content-type", "application/grpc" + content_type_suffix),
        ]
        if authority is not None:
            headers.insert(3, (":authority", authority))
        if timeout is not None:
            number_of_whole_seconds = timeout.days * 86400 + timeout.seconds
            if timeout.microseconds == 0:
                timeout_unit = "S"
                timeout_value = number_of_whole_seconds
            else:
                timeout_unit = "u"
                timeout_value = number_of_whole_seconds * 1000000 + timeout.microseconds
            timeout_str = "{}{}".format(timeout_value, timeout_unit)
            headers.insert(4, ("grpc-timeout", timeout_str))
        if message_type is not None:
            headers.append(("grpc-message-type", message_type))
        if self.config._message_encoding is not None:
            headers.append(("grpc-encoding", self.config._message_encoding))
        if self.config._message_accept_encoding is not None:
            headers.append(("grpc-accept-encoding", self.config._message_accept_encoding))
        if self.config._user_agent is not None:
            headers.append(("user-agent", self.config._user_agent))
        headers.extend(sanitize_headers(custom_metadata))
        self.h2_connection.send_headers(stream_id, headers, end_stream=False)

    def end_request(self, stream_id: int):
        self.h2_connection.send_data(stream_id, b"", end_stream=True)

    def start_response(self, stream_id: int, content_type_suffix="", custom_metadata=()):
        headers = [
            (":status", "200"),
            ("content-type", "application/grpc" + content_type_suffix),
        ]
        if self.config._message_encoding is not None:
            headers.append(("grpc-encoding", self.config._message_encoding))
        if self.config._message_accept_encoding is not None:
            headers.append(("grpc-accept-encoding", self.config._message_accept_encoding))
        headers.extend(sanitize_headers(custom_metadata))
        self.h2_connection.send_headers(stream_id, headers, end_stream=False)

    def respond_status(self, stream_id: int, status: Status, content_type_suffix="",
                       custom_metadata=()):
        trailers = [
            (":status", "200"),
            ("content-type", "application/grpc" + content_type_suffix),
            ("grpc-status", str(status.int_value)),
        ]
        if status.status_message:
            # TODO: should be percent encoded
            trailers.append(("grpc-message", status.status_message))
        trailers.extend(sanitize_headers(custom_metadata))
        self.h2_connection.send_headers(stream_id, trailers, end_stream=True)

    def end_response(self, stream_id: int, status: Status, custom_metadata=()):
        trailers = [
            ("grpc-status", str(status.int_value)),
        ]
        if status.status_message:
            # TODO: should be percent encoded
            trailers.append(("grpc-message", status.status_message))
        trailers.extend(sanitize_headers(custom_metadata))
        self.h2_connection.send_headers(stream_id, trailers, end_stream=True)



from .anyio_monkeypatch import apply_monkeypatch as _apply_anyio_monkeypatch

_apply_anyio_monkeypatch()


from purerpc.client import insecure_channel, Client
from purerpc.server import Service, Servicer, Server
from purerpc.rpc import Cardinality, RPCSignature, Stream
from purerpc.grpclib.status import Status, StatusCode
from purerpc.grpclib.exceptions import *



import re
import os
from setuptools import setup, find_packages

exec(open("src/purerpc/_version.py", encoding="utf-8").read())


def read(*names, **kwargs):
    with open(os.path.join(os.path.dirname(__file__), *names), "r") as fin:
        return fin.read()


def main():
    console_scripts = ['protoc-gen-purerpc=purerpc.protoc_plugin.plugin:main']
    setup(
        name="purerpc",
        version=__version__,
        license="Apache License Version 2.0",
        description="Asynchronous pure python gRPC server and client implementation using curio "
                    "and hyper-h2.",
        long_description='%s\n%s' % (
            re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.md')),
            re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('RELEASE.md'))
        ),
        long_description_content_type='text/markdown',
        author="Andrew Stepanov",
        url="https://github.com/standy66/purerpc",
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: POSIX :: Linux",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: Implementation :: CPython",
            "Programming Language :: Python :: Implementation :: PyPy",
            "Topic :: Internet",
            "Topic :: Software Development :: Libraries",
            "Topic :: System :: Networking",
        ],
        keywords=[
            "async", "await", "grpc", "pure-python", "pypy", "network",
            "rpc", "http2",
        ],
        packages=find_packages('src'),
        package_dir={'': 'src'},
        test_suite="tests",
        python_requires=">=3.5",
        install_requires=[
            "h2>=3.1.0,<4",
            "protobuf~=3.6",
            "anyio>=1.0.0b1,<2",
            "async_exit_stack>=1.0.1,<2",
            "tblib>=1.3.2,<2",
            "async_generator>=1.10,<2.0"
        ],
        entry_points={'console_scripts': console_scripts},
        setup_requires=["pytest-runner"],
        tests_require=[
            "pytest",
            "pytest-xdist",
            "pytest-forked",
            "grpcio_tools",
            "grpcio"
        ]
    )


if __name__ == "__main__":
    main()



import sys
import socket
import datetime

import anyio
from async_generator import async_generator, yield_, yield_from_
import h2
import h2.events
import h2.exceptions
from purerpc.utils import is_darwin
from purerpc.grpclib.exceptions import ProtocolError

from .grpclib.connection import GRPCConfiguration, GRPCConnection
from .grpclib.events import RequestReceived, RequestEnded, ResponseEnded, MessageReceived
from .grpclib.buffers import MessageWriteBuffer, MessageReadBuffer


class SocketWrapper:
    def __init__(self, grpc_connection: GRPCConnection, sock: anyio._networking.SocketStream):
        self._set_raw_socket_options(sock._socket._raw_socket)
        self._socket = sock._socket
        self._grpc_connection = grpc_connection
        self._write_fifo_lock = anyio.create_lock()
        self._retry = False

    @staticmethod
    def _set_raw_socket_options(raw_socket):
        raw_socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        if hasattr(socket, "TCP_KEEPIDLE"):
            raw_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 300)
        elif is_darwin():
            # Darwin specific option
            TCP_KEEPALIVE = 16
            raw_socket.setsockopt(socket.IPPROTO_TCP, TCP_KEEPALIVE, 300)
        raw_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 30)
        raw_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 5)
        raw_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

    async def flush(self):
        """This maybe called from different threads."""
        async with self._write_fifo_lock:
            self._retry = True
            while self._retry:
                self._retry = False
                data = self._grpc_connection.data_to_send()
                if not data:
                    return
                await self._socket.sendall(data)

    async def try_flush(self):
        if self.locked:
            self._retry = True
        else:
            await self.flush()

    @property
    def locked(self):
        return self._write_fifo_lock.locked()

    async def recv(self, buffer_size: int):
        """This may only be called from single thread."""
        return await self._socket.recv(buffer_size)


class GRPCStream:
    def __init__(self, grpc_connection: GRPCConnection, stream_id: int, socket: SocketWrapper,
                 grpc_socket: "GRPCSocket"):
        self._stream_id = stream_id
        self._grpc_connection = grpc_connection
        self._grpc_socket = grpc_socket
        self._socket = socket
        self._flow_control_update_event = anyio.create_event()
        self._incoming_events = anyio.create_queue(sys.maxsize)

        self._response_started = False
        self._closed = False
        
        self._start_stream_event = None
        self._end_stream_event = None

    @property
    def start_stream_event(self):
        return self._start_stream_event

    @property
    def end_stream_event(self):
        return self._end_stream_event

    @property
    def stream_id(self):
        return self._stream_id

    @property
    def client_side(self):
        return self._grpc_connection.config.client_side

    @property
    def debug_prefix(self):
        return "[CLIENT] " if self.client_side else "[SERVER] "

    async def _set_flow_control_update(self):
        await self._flow_control_update_event.set()

    async def _wait_flow_control_update(self):
        await self._flow_control_update_event.wait()
        self._flow_control_update_event.clear()

    async def _send(self, message: bytes, compress=False):
        message_write_buffer = MessageWriteBuffer(self._grpc_connection.config.message_encoding,
                                                  self._grpc_connection.config.max_message_length)
        message_write_buffer.write_message(message, compress)
        while message_write_buffer:
            window_size = self._grpc_connection.flow_control_window(self._stream_id)
            if window_size <= 0:
                await self._wait_flow_control_update()
                continue
            num_data_to_send = min(window_size, len(message_write_buffer))
            data = message_write_buffer.data_to_send(num_data_to_send)
            self._grpc_connection.send_data(self._stream_id, data)
            await self._socket.flush()

    async def _receive(self):
        event = await self._incoming_events.get()
        if isinstance(event, MessageReceived):
            self._grpc_connection.acknowledge_received_data(self._stream_id,
                                                            event.flow_controlled_length)
            await self._socket.flush()
        elif isinstance(event, RequestEnded) or isinstance(event, ResponseEnded):
            assert self._end_stream_event is None
            self._end_stream_event = event
        else:
            assert self._start_stream_event is None
            self._start_stream_event = event
        return event

    async def close(self, status=None, content_type_suffix="", custom_metadata=()):
        if self.client_side and (status or custom_metadata):
            raise ValueError("Client side streams cannot be closed with non-default arguments")
        if self._closed:
            raise TypeError("Closing already closed stream")
        self._closed = True
        if self.client_side:
            try:
                self._grpc_connection.end_request(self._stream_id)
            except h2.exceptions.StreamClosedError:
                # Remote end already closed connection, do nothing here
                pass
        elif self._response_started:
            self._grpc_connection.end_response(self._stream_id, status, custom_metadata)
        else:
            self._grpc_connection.respond_status(self._stream_id, status,
                                                 content_type_suffix, custom_metadata)
        await self._socket.flush()
        self._grpc_socket._decref_stream(self._stream_id)

    async def start_response(self, content_type_suffix="", custom_metadata=()):
        if self.client_side:
            raise ValueError("Cannot start response on client-side socket")
        self._grpc_connection.start_response(self._stream_id, content_type_suffix, custom_metadata)
        self._response_started = True
        await self._socket.flush()


# TODO: this name is not correct, should be something like GRPCConnection (but this name is already
# occupied)
class GRPCSocket:
    StreamClass = GRPCStream

    def __init__(self, config: GRPCConfiguration, sock,
                 receive_buffer_size=16384):
        self._grpc_connection = GRPCConnection(config=config)
        self._socket = SocketWrapper(self._grpc_connection, sock)
        self._receive_buffer_size = receive_buffer_size

        self._streams = {}  # type: Dict[int, GRPCStream]
        self._streams_count = {}  # type: Dict[int, int]

    @property
    def client_side(self):
        return self._grpc_connection.config.client_side

    def _stream_ctor(self, stream_id):
        return self.StreamClass(self._grpc_connection, stream_id, self._socket, self)

    def _allocate_stream(self, stream_id):
        stream = self._stream_ctor(stream_id)
        self._streams[stream_id] = stream
        self._streams_count[stream_id] = 2
        return stream

    def _decref_stream(self, stream_id: int):
        self._streams_count[stream_id] -= 1
        if self._streams_count[stream_id] == 0:
            del self._streams[stream_id]
            del self._streams_count[stream_id]

    @async_generator
    async def _listen(self):
        while True:
            data = await self._socket.recv(self._receive_buffer_size)
            if not data:
                return
            events = self._grpc_connection.receive_data(data)
            for event in events:
                # TODO: implement this
                if isinstance(event, h2.events.WindowUpdated):
                    if event.stream_id == 0:
                        for stream in self._streams.values():
                            await stream._set_flow_control_update()
                    elif event.stream_id in self._streams:
                        await self._streams[event.stream_id]._set_flow_control_update()
                    continue
                elif isinstance(event, RequestReceived):
                    self._allocate_stream(event.stream_id)

                await self._streams[event.stream_id]._incoming_events.put(event)

                if isinstance(event, RequestReceived):
                    await yield_(self._streams[event.stream_id])
                elif isinstance(event, ResponseEnded) or isinstance(event, RequestEnded):
                    self._decref_stream(event.stream_id)
            await self._socket.try_flush()

    async def _listener_thread(self):
        async for _ in self._listen():
            raise ProtocolError("Received request on client end")

    async def initiate_connection(self, parent_task_group: anyio.abc.TaskGroup):
        self._grpc_connection.initiate_connection()
        await self._socket.flush()
        if self.client_side:
            await parent_task_group.spawn(self._listener_thread)

    @async_generator
    async def listen(self):
        if self.client_side:
            raise ValueError("Cannot listen client-side socket")
        await yield_from_(self._listen())

    async def start_request(self, scheme: str, service_name: str, method_name: str,
                            message_type=None, authority=None, timeout: datetime.timedelta=None,
                            content_type_suffix="", custom_metadata=()):
        if not self.client_side:
            raise ValueError("Cannot start request on server-side socket")
        stream_id = self._grpc_connection.get_next_available_stream_id()
        stream = self._allocate_stream(stream_id)
        self._grpc_connection.start_request(stream_id, scheme, service_name, method_name,
                                            message_type, authority, timeout,
                                            content_type_suffix, custom_metadata)
        await self._socket.flush()
        return stream



import unittest
import functools
import collections
import subprocess
import multiprocessing
import tempfile
import shutil
import os
import sys
import inspect
import importlib
import concurrent.futures
import logging
import contextlib
import time
import random
import string
from queue import Empty as QueueEmpty

from tblib import pickling_support
pickling_support.install()

import anyio
from async_generator import aclosing


WrappedResult = collections.namedtuple("WrappedResult", ("result", "exc_info"))


def wrap_gen_in_process(queue):
    def decorator(gen):
        @functools.wraps(gen)
        def new_func(*args, **kwargs):
            try:
                for elem in gen(*args, **kwargs):
                    queue.put(WrappedResult(result=elem, exc_info=None))
            except:
                queue.put(WrappedResult(result=None, exc_info=sys.exc_info()))
            finally:
                queue.close()
                queue.join_thread()
        return new_func
    return decorator


class PureRPCTestCase(unittest.TestCase):
    @staticmethod
    async def async_iterable_to_list(async_iterable):
        result = []
        async with aclosing(async_iterable) as async_iterable:
            async for value in async_iterable:
                result.append(value)
        return result

    @staticmethod
    def random_payload(min_size=1000, max_size=100000):
        return "".join(random.choice(string.ascii_letters)
                       for _ in range(random.randint(min_size, max_size)))

    @staticmethod
    @contextlib.contextmanager
    def run_context_manager_generator_in_process(cm_gen):
        queue = multiprocessing.Queue()
        target_fn = wrap_gen_in_process(queue)(cm_gen)

        process = multiprocessing.Process(target=target_fn)
        process.start()
        try:
            wrapped_result = queue.get()
            if wrapped_result.exc_info is not None:
                raise wrapped_result.exc_info[0].with_traceback(*wrapped_result.exc_info[1:])
            else:
                yield wrapped_result.result
        finally:
            try:
                exc_info = queue.get_nowait().exc_info
                if exc_info is not None:
                    raise exc_info[0].with_traceback(*exc_info[1:])
            except QueueEmpty:
                pass
            finally:
                process.terminate()
                process.join()
                queue.close()
                queue.join_thread()

    @classmethod
    def run_purerpc_service_in_process(self, service):
        def target_fn():
            import purerpc
            server = purerpc.Server(port=0)
            server.add_service(service)
            socket = server._create_socket_and_listen()
            yield socket.getsockname()[1]
            anyio.run(server._run_async_server, socket)
        return self.run_context_manager_generator_in_process(target_fn)

    @classmethod
    def run_grpc_service_in_process(self, add_handler_fn):
        def target_fn():
            import grpc
            server = grpc.server(concurrent.futures.ThreadPoolExecutor(max_workers=8))
            port = server.add_insecure_port('[::]:0')
            add_handler_fn(server)
            server.start()
            yield port
            while True:
                time.sleep(60)
        return self.run_context_manager_generator_in_process(target_fn)

    @staticmethod
    def run_tests_in_workers(*, target, num_workers):
        queue = multiprocessing.Queue()

        @wrap_gen_in_process(queue)
        def target_fn():
            target()
            yield

        processes = [multiprocessing.Process(target=target_fn) for _ in range(num_workers)]
        for process in processes:
            process.start()

        try:
            for _ in range(num_workers):
                wrapped_result = queue.get()
                if wrapped_result.exc_info is not None:
                    raise wrapped_result.exc_info[0].with_traceback(*wrapped_result.exc_info[1:])
        finally:
            queue.close()
            queue.join_thread()
            for process in processes:
                process.join()

    @staticmethod
    @contextlib.contextmanager
    def compile_temp_proto(*relative_proto_paths):
        modules = []
        with tempfile.TemporaryDirectory() as temp_dir:
            sys.path.insert(0, temp_dir)
            try:
                for relative_proto_path in relative_proto_paths:
                    proto_path = os.path.join(os.path.dirname(
                        inspect.currentframe().f_back.f_back.f_globals['__file__']),
                        relative_proto_path)
                    proto_filename = os.path.basename(proto_path)
                    proto_temp_path = os.path.join(temp_dir, proto_filename)
                    shutil.copyfile(proto_path, proto_temp_path)
                for relative_proto_path in relative_proto_paths:
                    proto_filename = os.path.basename(relative_proto_path)
                    proto_temp_path = os.path.join(temp_dir, proto_filename)
                    cmdline = [sys.executable, '-m', 'grpc_tools.protoc', '--python_out=.',
                               '--purerpc_out=.', '-I' + temp_dir, proto_temp_path]
                    subprocess.check_call(cmdline, cwd=temp_dir)
                    pb2_module_name = proto_filename.replace(".proto", "_pb2")
                    grpc_module_name = proto_filename.replace(".proto", "_grpc")
                    pb2_module = importlib.import_module(pb2_module_name)
                    grpc_module = importlib.import_module(grpc_module_name)
                    modules.extend((pb2_module, grpc_module))
                yield modules
            finally:
                sys.path.remove(temp_dir)

    def setUp(self):
        self.configure_logs()

    @staticmethod
    def configure_logs():
        logging.basicConfig(format="[%(asctime)s - %(name)s - %(levelname)s]:  %(message)s", level=logging.WARNING)
